local B, L, T = unpack(select(2, ...))

T.LABEL = {
    ['bank'] = L[10001],
    ['reagent'] = L[10002],
    ['void'] = L[10003],
    ['guild'] = L[10004]
}

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML bankswatch.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function BW_OnLoad(self)

        ----- REGISTER FOR EVENTS
            self:RegisterEvent('BANKFRAME_OPENED')
            self:RegisterEvent('BANKFRAME_CLOSED')
            self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
            self:RegisterEvent('PLAYERBANKBAGSLOTS_CHANGED')
            self:RegisterEvent('PLAYERREAGENTBANKSLOTS_CHANGED')
            self:RegisterEvent('VOID_STORAGE_OPEN')
            self:RegisterEvent('VOID_STORAGE_CLOSE')
            self:RegisterEvent('VOID_STORAGE_UPDATE')
            self:RegisterEvent('VOID_STORAGE_CONTENTS_UPDATE')
            self:RegisterEvent('VOID_TRANSFER_DONE')
            self:RegisterEvent('GUILDBANKFRAME_OPENED')
            self:RegisterEvent('GUILDBANKFRAME_CLOSED')
            self:RegisterEvent('GUILDBANKBAGSLOTS_CHANGED')
            self:RegisterEvent('GUILDBANK_UPDATE_TABS')
            self:RegisterEvent('PLAYER_ENTERING_WORLD')
            self:RegisterEvent('VARIABLES_LOADED')
            self:RegisterEvent('PLAYER_REGEN_DISABLED')
            self:RegisterEvent('GET_ITEM_INFO_RECEIVED')

        DEFAULT_CHAT_FRAME:AddMessage(T.MSGC.DEFAULT..L[10000])

        ------ SLASH COMMANDS
            SLASH_BW1 = "/BW"
            SLASH_BW2 = "/bw"
            SLASH_BW3 = "/Bw"
            SLASH_BW4 = "/bW"
            SLASH_BW5 = "/Bankswatch"
            SLASH_BW6 = "/BanksWatch"
            SLASH_BW7 = "/BANKSWATCH"
            SLASH_BW8 = "/bankswatch"

            SlashCmdList['BW'] = B.Main

            B.CreateFrames()
            table.insert(UISpecialFrames, T.F.mainframe.bank:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.reagent:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.void:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.guild:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.search:GetName())
    end

    function BW_OnEvent(self, event, ...)
        B.print(event)

        if event == "VARIABLES_LOADED" then
            B.print(T.MSGC.EVENT.."Variables loaded. Initializing.", 2)
            B.print(T.MSGC.INFO.."Running version: "..GetAddOnMetadata("BanksWatch", "Version"), 1)

            if BWsavedtable then
                T.BUTTON_HEIGHT = BWsavedtable['iconsize'] or 32
                T.ROW = BWsavedtable['iconrow'] or 12

                if BWsavedtable['minimize'] then T.F.mainframe.togglebutton:Show() end
            end

            B.I.Main.Bank.Buttons()
            B.I.Main.Reagent.Buttons()
            B.I.Main.Void.Buttons()
            B.I.Main.Guild.Buttons()

        elseif event == "GET_ITEM_INFO_RECEIVED" then
            local itemid = ...
            if T.BW.lookup[itemid] then T.BW.lookup[itemid] = nil end

            if T.BW.lookuprequest and not next(T.BW.lookup) then -- Try to minimize the amount of updates
                T.BW.lookuprequest = false
                B.BankUpdate()
                B.ReagentUpdate()
                B.VoidUpdate()
                B.GuildUpdate()
            end

        elseif event == "PLAYER_ENTERING_WORLD" then
            -- When loading, BAG_UPDATE triggers, but I only want when player modifies their bank bags
            self:RegisterEvent("PLAYER_REGEN_DISABLED")
            self:RegisterEvent("BAG_UPDATE")

            B.Initialize() -- Load or create character
            B.C.Create() -- Config
            B.SetFrameLocation()

        elseif event == "PLAYERBANKSLOTS_CHANGED" then
            -- Only for generic bank bag
            B.BankSlotScan(...)

        elseif event == "BAG_UPDATE" then
            local bagid = ...

            if bagid == -1 or (bagid >= (NUM_BAG_SLOTS + 1) and bagid <= (NUM_BAG_SLOTS + NUM_BANKBAGSLOTS)) then
                -- bank bags
                B.BankBagScan(...)
            end

        elseif event == "PLAYERREAGENTBANKSLOTS_CHANGED" then
            B.ReagentScanSlot(...)

        elseif event == "VOID_STORAGE_OPEN" or event == "VOID_STORAGE_UPDATE" or event == "VOID_STORAGE_CONTENTS_UPDATE" or event == "VOID_TRANSFER_DONE" then
            B.VoidScan()

        elseif event == "VOID_STORAGE_CLOSE" then
            B.VoidUpdate()

        elseif event == "BANKFRAME_OPENED" then
            T.ISBANKOPEN = true -- Prevent BAG_UPDATE from taking effect with first login

            B.HideAll()
            B.BankScan()
            B.ReagentScan()

        elseif event == "BANKFRAME_CLOSED" then
            T.ISBANKOPEN = false

            B.BankUpdate()
            B.ReagentUpdate()

        elseif event == "GUILDBANKFRAME_OPENED" or event == "GUILDBANK_UPDATE_TABS" then
            B.RefreshGuildTab()

        elseif event == "GUILDBANKBAGSLOTS_CHANGED" then
            -- Guild can change outside the scope of the player. Refresh everytime the guild bank is opened.
            -- This will trigger when information about tabs is found.
            B.GuildScan()

        elseif event == "GUILDBANKFRAME_CLOSED" then
            B.GuildUpdate()

        elseif event == "PLAYER_REGEN_DISABLED" then
            -- Entering combat, hide any open bags
            B.HideAll()
        end
    end

--[[//////////////////////////////////////////////////

    MAIN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Initialize()

        hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", B.ContainerFrameItemButton_OnModifiedClick)
        hooksecurefunc("ChatFrame_OnHyperlinkShow", B.ChatFrame_OnHyperlinkShow)
        hooksecurefunc("ChatEdit_InsertLink", B.InsertLink)

        local player = UnitName('player')
        local server = GetRealmName()
        local faction = UnitFactionGroup('player')

        if (player == nil) or (player == UNKNOWNOBJECT) or (player == UNKNOWNBEING) then
            return
        end

        if BWsavedtable and BWsavedtable[server] and BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] then
            B.load(server, player, faction)
        else
            B.print(T.MSGC.EVENT.."New character detected")
            B.template(server, player, faction)
        end
    end

    function B.Main(input)

        B.print(T.MSGC.INFO.."That's a lot of stuff.", 1)
        if input then input = string.lower(input) end

        if input == "toggle" then
            if T.F.toggleframe:IsVisible() then
                T.F.toggleframe:Hide()
            else
                T.F.toggleframe:Show()
            end

        elseif input == "debug" then
            BWdebug = not BWdebug
            B.print(T.MSGC.INFO.."Debug:|r "..T.MSGC.BOOL..tostring(BWdebug), 1)

        elseif input == "tooltip" then
            BWsavedtable.helptooltip = not BWsavedtable.helptooltip
            B.print(T.MSGC.INFO.."Help Tooltip:|r "..T.MSGC.BOOL..tostring(BWsavedtable.helptooltip), 1)
        end
    end

    function B.ContainerFrameItemButton_OnModifiedClick(self)

        if IsShiftKeyDown() then
            local bag, item = self:GetParent():GetID(), self:GetID()
            local link = GetContainerItemLink(bag, item)

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then

                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.ChatFrame_OnHyperlinkShow(self, link, text, button)

        if IsShiftKeyDown() and link then
            if string.find(link, 'achievement:*') or string.find(link,'spell:*') then
                return false
            end

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.InsertLink(text)

        if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
            T.F.mainframe.searchbar:SetText(string.find(text, "^|c%x+|H(.+)|h%[.*%]"))
        end
    end

    function B.dropDownMenuItem_OnClick(self)
        local server, player, faction
        local value = self.value

        server, player = value:match"([^.]*).(.*)"
        player, faction = player:match"([^.]*).(.*)"
        faction = faction == "H" and "Horde" or faction == "A" and "Alliance" or faction == "N" and "Neutral"

        B.load(server, player, faction)
        BWdropDownMenuButton:Click()
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.BankScan()

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, counter, count, link, bag, slot, slots
        local bank = B.GetPlayerBank('bank', true)

        if not bank[1] then bank[1] = B.newtable() end
        for x = 1, NUM_BANKGENERIC_SLOTS do
            _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, x)

            if link then
                bank[1][x] = B.newtable()
                bank[1][x]['count'] = count
                bank[1][x]['link'] = link
            else
                bank[1][x] = B.deltable(bank[1][x])
            end
        end

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            if not bank[bag] then bank[bag] = B.newtable() end
            slots = GetContainerNumSlots(bag)
            --[[local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
            local baglink = GetInventoryItemLink("player", inventoryID)
            B.print(baglink)]]
            bank[bag]['slots'] = slots

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    bank[bag][x] = B.newtable()
                    bank[bag][x]['link'] = link
                    bank[bag][x]['count'] = count > 1 and count or nil
                else
                    bank[bag][x] = B.deltable(bank[bag][x])
                end
            end
        end

        bank = nil
        collectgarbage()
    end

    function B.BankSlotScan(slot)

        local bank = B.GetPlayerBank('bank', true)
        local _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, slot)

        if link then
            bank[1][slot] = B.newtable()
            bank[1][slot]['link'] = link
            bank[1][slot]['count'] = count > 1 and count or nil
        else
            bank[1][slot] = B.deltable(bank[1][slot])
        end
    end

    function B.BankBagScan(bag)

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, slots, count, link
        local bank = B.GetPlayerBank('bank', true)
        if not bank[bag] then bank[bag] = B.newtable() end

        slots = GetContainerNumSlots(bag)
        bank[bag]['slots'] = slots

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

            if link then
                bank[bag][x] = B.newtable()
                bank[bag][x]['link'] = link
                bank[bag][x]['count'] = count > 1 and count or nil
            else
                bank[bag][x] = B.deltable(bank[bag][x])
            end
        end

        bank = nil
        collectgarbage()
    end

    function B.ReagentScan()

        local x,  page, count, link
        local reagent = B.GetPlayerBank('reagent', true)

        for page = 1, T.REAGENT_PAGES do
            if not reagent[page] then reagent[page] = B.newtable() end

            for x = 1, T.NUM_REAGENT_SLOTS do
                _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, x)

                if link then
                    reagent[page][x] = B.newtable()
                    reagent[page][x]['link'] = link
                    reagent[page][x]['count'] = count > 1 and count or nil
                else
                    reagent[page][x] = B.deltable(reagent[page][x])
                end
            end
        end

        reagent = nil
        collectgarbage()
    end

    function B.ReagentScanSlot(slot)

        local reagent = B.GetPlayerBank('reagent', true)
        local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

        if link then
            reagent[1][slot] = B.newtable()
            reagent[1][slot]['link'] = link
            reagent[1][slot]['count'] = count > 1 and count or nil
        else
            reagent[1][slot] = B.deltable(reagent[1][slot])
        end
    end

    function B.VoidScan()

        if not IsVoidStorageReady() then return end
        local x, page, itemID
        local void = B.GetPlayerBank('void', true)

        for page = 1, T.VOID_PAGES do
            if not void[page] then void[page] = B.newtable() end

            for x = 1, T.NUM_VOID_SLOTS do
                itemID = GetVoidItemInfo(page, x)

                if itemID then
                    void[page][x] = B.newtable()
                    void[page][x]['link'] = page == 1 and GetVoidItemHyperlinkString(x) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + x)
                else
                    void[page][x] = B.deltable(void[page][x])
                end
            end
        end

        void = nil
        collectgarbage()
    end

    function B.GuildScan()

        local x, i, id, count, item, button
        local guild = B.GetPlayerBank('guild', true)
        local page = GetCurrentGuildBankTab()

        if guild[page].access then
            for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                _, count = GetGuildBankItemInfo(page, x)
                if count > 0 then
                    guild[page][x] = B.newtable()
                    guild[page][x]['link'] = GetGuildBankItemLink(page, x)
                    guild[page][x]['count'] = count > 1 and count or nil
                else
                    guild[page][x] = B.deltable(guild[page][x])
                end
            end
            --[[for x = 1, NUM_GUILDBANK_COLUMNS do
                for i = 1, NUM_SLOTS_PER_GUILDBANK_GROUP do

                    button = _G["GuildBankColumn"..x.."Button"..i]
                    count = tonumber(button.count)
                    id = button:GetID()

                    if count > 0 then
                        guild[page][id] = B.newtable()
                        guild[page][id]['link'] = GetGuildBankItemLink(page, id)
                        guild[page][id]['count'] = count > 1 and count or nil
                    else
                        guild[page][id] = B.deltable(guild[page][id])
                    end
                end
            end]]
        end

        guild = nil
        collectgarbage()
    end

--[[//////////////////////////////////////////////////

    UPDATE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.ScrollbarUpdate()
        local r, g, b, idx, link, hexcolor, itemRarity
        local total_items = #T.BW.items
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.search.scroll)

        FauxScrollFrame_Update(T.F.mainframe.search.scroll, total_items, B.RowsToFit(), T.BUTTON_HEIGHT)
        
        for x = 1, B.RowsToFit() do
            -- Get the appropriate item, which will be x + value
            idx = x + currentscrollbarvalue

            B.ResetButton(T.F.mainframe.search.list.buttons[x])
            T.F.mainframe.search.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.search.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.search.list.buttons[x]:Hide()
            T.F.mainframe.search.enddot:Hide()

            if T.BW.items[idx] then
                r, g, b, hexcolor = GetItemQualityColor(T.BW.items[idx].quality)

                T.F.mainframe.search.list.buttons[x].quality = T.BW.items[idx].quality
                T.F.mainframe.search.list.buttons[x].icon:SetTexture(T.BW.items[idx].icon)
                T.F.mainframe.search.list.buttons[x].leftstring:SetText("|c"..hexcolor..T.BW.items[idx].name)

                if T.BW.items[idx].count > 1 then
                    T.F.mainframe.search.list.buttons[x].text:SetText(T.BW.items[idx].count)
                end
                if T.BW.items[idx].ilvl and T.BW.items[idx].ilvl ~= "" then
                    T.F.mainframe.search.list.buttons[x].ilvl:SetText(T.BW.items[idx].ilvl)
                    T.F.mainframe.search.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                end

                loc, num = T.BW.items[idx].loc:match"([^.]*).(.*)"
                location = tonumber(num) == 1 and T.LABEL[loc] or T.LABEL[loc].." "..num
                T.F.mainframe.search.list.buttons[x].location = {['loc'] = loc, ['num'] = num}
                T.F.mainframe.search.list.buttons[x].rightstring:SetText(location)
                T.F.mainframe.search.list.buttons[x].link = T.BW.items[idx].link
                T.F.mainframe.search.list.buttons[x]:Show()
                T.F.mainframe.search.enddot:Show()
                if idx == total_items then
                    T.F.mainframe.search.enddot:Hide()
                end
            end
        end
    end

    function B.BankUpdate()

        local page, total, counter
        local bank = B.GetPlayerBank('bank')

        total = 0
        for page = 1, #bank do
            total = total + (bank[page] and bank[page]['slots'] or 0)
        end
        B.ValidateBankBags(total)

        counter = 0
        for page = 1, #bank do
            if bank[page] then

                B.UpdatePage(T.F.mainframe.bank, bank[page]['slots'], bank[page], counter)
                counter = counter + bank[page]['slots']
            end
        end
    end

    function B.ReagentUpdate()

        local page
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            B.UpdatePage(T.F.mainframe.reagent[page], T.NUM_REAGENT_SLOTS, reagent[page])
        end
    end

    function B.VoidUpdate()

        local page
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do
            B.UpdatePage(T.F.mainframe.void[page], T.NUM_VOID_SLOTS, void[page])
        end
    end

    function B.GuildUpdate()

        local page
        local guild = B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do
            if guild[page] then

                if guild[page].icon then
                    T.F.mainframe.guild[page].tab:SetNormalTexture(guild[page].icon)
                    T.F.mainframe.guild[page].tab.name = guild[page].name
                    T.F.mainframe.guild[page].tab:Show()
                else
                    T.F.mainframe.guild[page].tab:Hide()
                end

                if guild[page].access then -- Validate user has access to view tab
                    T.F.mainframe.guild[page].tab:Enable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(false)
                    B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, guild[page])
                else
                    T.F.mainframe.guild[page].tab:Disable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(true)
                end
            else
                T.F.mainframe.guild[page].tab:Hide()
            end
        end
    end

    function B.RefreshGuildTab()

        local x, name, icon, isViewable
        local guild = B.GetPlayerBank('guild')
        local tabs = GetNumGuildBankTabs()

        for x = 1, MAX_GUILDBANK_TABS do
            if x <= tabs then
                if not guild[x] then guild[x] = B.newtable() end
                name, icon, isViewable = GetGuildBankTabInfo(x)
                guild[x]['icon'] = icon
                guild[x]['name'] = name
                guild[x]['access'] = isViewable

                T.F.mainframe.guild[x].tab:SetNormalTexture(icon)
                T.F.mainframe.guild[x].tab.name = name
                T.F.mainframe.guild[x].tab:Show()
            else
                guild[x] = B.deltable(guild[x])
                T.F.mainframe.guild[x].tab:Hide()
            end
        end
    end

    function B.UpdatePage(page, maxslots, data, optstart)
        local x, item
        optstart = optstart or 0

        for x = 1, maxslots do

            B.ResetButton(page.buttons[x + optstart])
            if data and data[x] then

                item = {GetItemInfo(data[x].link)}
                if not next(item) then -- Item is not cached yet
                    T.BW.lookuprequest = true
                    T.BW.lookup[tonumber(B.GetItemIdFromLink(data[x].link))] = true
                else
                    B.UpdateButton(page.buttons[x + optstart], item, data[x])
                end
            end
        end
    end

    function B.UpdateButton(button, item, data)
        local r, g, b, ilvl

        r, g, b = GetItemQualityColor(item[3])

        button.name = item[1]
        button.count = data.count or 1
        button.quality = item[3]
        button.link = item[2]

        button.icon:SetTexture(item[10])
        button.icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        
        if (data.count or 1) > 1 then button.text:SetText(data.count) end
        if item[3] > 1 then button:SetBackdropBorderColor(r, g, b, 1) end

        if item[6] == BATTLE_PET_SOURCE_2 then -- QUEST
            button:SetBackdropBorderColor(1, 1, 0, 1)
            button.quest = true

        elseif item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON then -- ILVL
            ilvl = B.DetectILVL(item[2])
            if ilvl then
                button.ilvl:SetText(ilvl)
                button.ilvl:SetTextColor(r, g, b, 1)
            else
                button.ilvl:SetText(item[4])
                button.ilvl:SetTextColor(r, g, b, 1)
            end
        end
    end

    function B.ResetButton(button)

        button.icon:SetTexture("")
        button:SetBackdropBorderColor(0, 0, 0, 1)
        button.link = nil
        button.quality = nil
        button.quest = nil
        button.name = nil
        button.count = nil
        button.text:SetText("")
        button.ilvl:SetText("")
    end

--[[//////////////////////////////////////////////////

    SEARCH FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Search(text)

        T.BW.items = B.deltable(T.BW.items)
        T.BW.items = B.newtable()

        B.BankSearch(text)
        B.ReagentSearch(text)
        B.VoidSearch(text)
        B.GuildSearch(text)

        B.SortSearch()
        B.ScrollbarUpdate()

        if not T.F.mainframe.search:IsVisible() then T.F.mainframe.countbutton:Click() end

        collectgarbage()
    end

    function B.SortSearch()
        local banks = {"bank", "reagent", "void", "guild"}

        if not next(T.BW.items) then return end

        local sort = B.newtable()
        for x = 1, #banks do sort[banks[x]] = B.newtable() end

        -- sort results, group by location, stack same item
        for x = 1, #T.BW.items do
            name = T.BW.items[x].name
            loc, num = T.BW.items[x].loc:match"([^.]*).(.*)"
            num = tonumber(num)

            if not sort[loc][num] then sort[loc][num] = B.newtable() end
            if not sort[loc][num][name] then
                sort[loc][num][name] = B.newtable()
                sort[loc][num][name]['count'] = 0
                sort[loc][num][name]['link'] = T.BW.items[x].link
                sort[loc][num][name]['quality'] = T.BW.items[x].quality
                sort[loc][num][name]['icon'] = T.BW.items[x].icon
                sort[loc][num][name]['name'] = name
                sort[loc][num][name]['ilvl'] = T.BW.items[x].ivl
                sort[loc][num][name]['loc'] = T.BW.items[x].loc
            end
            sort[loc][num][name]['stack'] = B.newtable()
            sort[loc][num][name]['stack'][#sort[loc][num][name]['stack'] + 1] = B.newtable()
            B.copytable(sort[loc][num][name]['stack'][#sort[loc][num][name]['stack']], T.BW.items[x])
            sort[loc][num][name]['count'] = sort[loc][num][name]['count'] + T.BW.items[x].count
        end

        T.BW.items = B.deltable(T.BW.items)
        T.BW.items = B.newtable()

        for x = 1, #banks do
            for i = 1, table.maxn(sort[banks[x]]) do
                if sort[banks[x]][i] then
                    for k, v in pairs(sort[banks[x]][i]) do
                        T.BW.items[#T.BW.items + 1] = B.newtable()
                        B.copytable(T.BW.items[#T.BW.items], v)
                    end
                end
            end
        end

        sort = nil
    end

    function B.BankSearch(text)

        local x, page, counter
        local bank = B.GetPlayerBank('bank')
        if not bank or next(bank) == nil then
            B.print(T.MSGC.ERROR.."Bank is empty")
            return false
        end

        x = 1
        for page = 1, #bank do
            if bank[page] then
                for counter = 1, bank[page]['slots'] do

                    T.F.mainframe.bank.buttons[x].search:Hide()
                    name = T.F.mainframe.bank.buttons[x].name

                    if name and text then
                        if B.find(name, text) then
                            T.F.mainframe.bank.buttons[x].loc = "bank."..(page == 1 and page or page - 4)
                            B.SearchSlot(T.F.mainframe.bank.buttons[x])
                        else
                            T.F.mainframe.bank.buttons[x].search:Show()
                        end
                    end
                    x = x + 1
                end
            end
        end
    end

    function B.ReagentSearch(text)

        local x, page
        local reagent = B.GetPlayerBank('reagent')
        if not reagent or next(reagent) == nil then
            B.print(T.MSGC.ERROR.."Reagent bank is empty")
            return false
        end

        for page = 1, T.REAGENT_PAGES do
            for x = 1, T.NUM_REAGENT_SLOTS do

                T.F.mainframe.reagent[page].buttons[x].search:Hide()
                name = T.F.mainframe.reagent[page].buttons[x].name

                if name and text then
                    if B.find(name, text) then
                        T.F.mainframe.reagent[page].buttons[x].loc = "reagent."..page
                        B.SearchSlot(T.F.mainframe.reagent[page].buttons[x])
                    else
                        T.F.mainframe.reagent[page].buttons[x].search:Show()
                    end
                end
            end
        end
    end

    function B.VoidSearch(text)

        local x, proceed, page
        local void = B.GetPlayerBank('void')
        
        proceed = false
        for page = 1, T.VOID_PAGES do
            if void[page] and next(void[page]) then
                proceed = true
                break
            end
        end
        if not proceed then
            B.print(T.MSGC.ERROR.."Void storage is empty")
            return false
        end

        for page = 1, T.VOID_PAGES do
            for x = 1, T.NUM_VOID_SLOTS do

                T.F.mainframe.void[page].buttons[x].search:Hide()
                name = T.F.mainframe.void[page].buttons[x].name

                if name and text then
                    if B.find(name, text) then
                        T.F.mainframe.void[page].buttons[x].loc = "void."..page
                        B.SearchSlot(T.F.mainframe.void[page].buttons[x])
                    else
                        T.F.mainframe.void[page].buttons[x].search:Show()
                    end
                end
            end
        end
    end

    function B.GuildSearch(text)

        local x, proceed, page
        local guild = B.GetPlayerBank('guild')

        proceed = false
        for page = 1, MAX_GUILDBANK_TABS do
            if guild[page] and next(guild[page]) then
                proceed = true
                break
            end
        end
        if not proceed then
            B.print(T.MSGC.ERROR.."Guild bank is empty")
            return false
        end

        for page = 1, MAX_GUILDBANK_TABS do

            if guild[page] and guild[page].access then
                --T.F.mainframe.guild[page].tab.search:Hide()

                for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                    T.F.mainframe.guild[page].buttons[x].search:Hide()
                    name = T.F.mainframe.guild[page].buttons[x].name

                    if name and text then
                        if B.find(name, text) then
                            T.F.mainframe.guild[page].buttons[x].loc = "guild."..page
                            B.SearchSlot(T.F.mainframe.guild[page].buttons[x])
                        else
                            T.F.mainframe.guild[page].buttons[x].search:Show()
                        end
                    end
                end
                --if not found then T.F.mainframe.guild[page].tab.search:Show() end
            end
        end
    end

    function B.SearchSlot(button)

        button.search:Hide()

        T.BW.items[#T.BW.items + 1] = B.newtable()
        T.BW.items[#T.BW.items]['link'] = button.link
        T.BW.items[#T.BW.items]['quality'] = button.quality
        T.BW.items[#T.BW.items]['name'] = button.name
        T.BW.items[#T.BW.items]['count'] = button.count
        T.BW.items[#T.BW.items]['quest'] = button.quest
        T.BW.items[#T.BW.items]['icon'] = button.icon:GetTexture()
        T.BW.items[#T.BW.items]['ilvl'] = button.ilvl:GetText()
        T.BW.items[#T.BW.items]['loc'] = button.loc

        T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + button.count
    end
