local B, L, T = unpack(select(2, ...))


B.I = {['bank'] = {['num'] = 0, ['rows'] = 0},
       ['reagent'] = {['num'] = 0, ['rows'] = 0},
       ['void'] = {},
       ['guild'] = {}}

--[[//////////////////////////////////////////////////

    MAIN INTERFACE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    B.I.Main = {}
    function B.CreateFrames()

        ----- Button bar with search
            B.I.Main.Frame.Create()

        ----- bank frame
            B.I.Main.Bank.Create()

        ------ reagent frame
            B.I.Main.Reagent.Create()

        ------ void frame
            B.I.Main.Void.Create()

        ------ guild frame
            B.I.Main.Guild.Create()

        ------ search frame
            B.I.Main.Search.Create()
    end

    function B.ValidateBankBags(slots)

        local i, button, bankrow
        local numslots = (T.ROW * B.I.bank.rows) + B.I.bank.num
        local lastbutton = T.F.mainframe.bank.buttons[numslots]
        local bankrow = T.F.mainframe.bank.buttons[T.ROW * B.I.bank.rows + 1]

        if not slots or slots == numslots then return-- Already correct amount of slots
        elseif slots < numslots then -- Hide extra slots
            local difference = numslots - slots
            for i = numslots, (numslots - difference + 1), -1 do

                T.F.mainframe.bank.buttons[i]:Hide()

                if B.I.bank.num == 1 then
                    B.I.bank.num = T.ROW
                    B.I.bank.rows = B.I.bank.rows - 1
                else
                    B.I.bank.num = B.I.bank.num - 1
                end
            end
        else -- Add or Show missing slots
            for i = numslots + 1, slots do

                if T.F.mainframe.bank.buttons[i] then
                    T.F.mainframe.bank.buttons[i]:Show()
                    if B.I.bank.num == T.ROW then
                        B.I.bank.rows = B.I.bank.rows + 1
                        B.I.bank.num = 1
                    else
                        B.I.bank.num = B.I.bank.num + 1
                    end
                    lastbutton = T.F.mainframe.bank.buttons[i]
                else
                    button = B.I.Button.Create(i, T.F.mainframe.bank.frame)
                    button:SetScript("OnEnter", B.I.Button.Enter)
                    button:SetScript("OnLeave", function(self) GameTooltip:Hide() end)

                    if B.I.bank.num == T.ROW then -- Last column of the row
                        button:SetPoint("TOP", bankrow, "BOTTOM", 0, -T.FRAMEWHITESPACE)
                        bankrow = button
                        B.I.bank.rows = B.I.bank.rows + 1
                        B.I.bank.num = 1
                    else
                        button:SetPoint("TOPLEFT", lastbutton, "TOPRIGHT", T.FRAMEWHITESPACE, 0)
                        B.I.bank.num = B.I.bank.num + 1
                    end
                    T.F.mainframe.bank.buttons[i] = button
                    lastbutton = button
                end
            end
        end
        T.F.mainframe.bank:SetHeight(10 + (T.BUTTON_HEIGHT * (B.I.bank.rows + 2)) + (T.FRAMEWHITESPACE * B.I.bank.rows))
        T.F.mainframe.bank:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
    end

    B.I.Main.Frame = {}
    function B.I.Main.Frame.Create()

        T.F.toggleframe = CreateFrame("Frame", "BWtoggleframe", UIParent)
        T.F.toggleframe:SetAllPoints()

        T.F.mainframe = CreateFrame("Frame", "BWmainframe", T.F.toggleframe)
        T.F.mainframe:SetPoint("CENTER")
        T.F.mainframe:SetHeight(25)
        T.F.mainframe:SetWidth(15 + (32 * 5) + 192) -- options + searchbar
        T.F.mainframe:SetFrameStrata("DIALOG")
        T.F.mainframe.layout = 1

        B.I.Backdrop(T.F.mainframe)
        T.F.mainframe:SetMovable(true)
        T.F.mainframe:SetUserPlaced(true)
        T.F.mainframe:SetClampedToScreen(true)

        T.F.mainframe:SetScript("OnMouseDown", B.I.Main.Frame.MouseDown)
        T.F.mainframe:SetScript("OnMouseUp", B.I.Main.Frame.MouseUp)

        -- Create option button
        B.I.Main.Frame.Options()

        -- Create bank button
        B.I.Main.Frame.Bank()

        -- Create reagent button
        B.I.Main.Frame.Reagent()

        -- Create void button
        B.I.Main.Frame.Void()

        -- Create guild button
        B.I.Main.Frame.Guild()

        -- Search bar
        B.I.Main.Frame.Search()

        -- Create count button
        B.I.Main.Frame.Count()
    end

    function B.I.Main.Frame.Button(label, parent, name)
        local button = CreateFrame("Button", name, parent or T.F.mainframe)

        button:SetWidth(30)
        button:SetHeight(T.F.mainframe:GetHeight())
        button:SetNormalFontObject("GameFontNormal")
        button:SetText(label)
        button:SetBackdrop({bgFile = "Interface\\Buttons\\WHITE8x8",  edgeFile = "Interface\\Buttons\\WHITE8x8", edgeSize = 1})
        button:SetBackdropBorderColor(0, 0, 0, 0)

        local highlight = button:CreateTexture(nil, "HIGHLIGHT")
        highlight:SetTexture("Interface\\Buttons\\WHITE8x8")
        highlight:SetHeight(2)
        highlight:SetVertexColor(1, 1, 1, 1)
        button:SetHighlightTexture(highlight)
        B.I.Main.Frame.ButtonHighlight(button, "BOTTOM")

        return button
    end

    function B.I.Main.Frame.ButtonHighlight(button, orientation)
        local highlight = button:GetHighlightTexture()

        highlight:ClearAllPoints()
        highlight:SetPoint(orientation.."LEFT", button, orientation.."LEFT")
        highlight:SetPoint(orientation.."RIGHT", button, orientation.."RIGHT")
    end

    function B.I.Main.Frame.ButtonOrder(orientation)

        T.F.mainframe.optionbutton:ClearAllPoints()
        T.F.mainframe.bankbutton:ClearAllPoints()
        T.F.mainframe.reagentbutton:ClearAllPoints()
        T.F.mainframe.voidbutton:ClearAllPoints()
        T.F.mainframe.guildbutton:ClearAllPoints()
        T.F.mainframe.searchbar:ClearAllPoints()
        T.F.mainframe.countbutton:ClearAllPoints()
        T.F.mainframe.search:ClearAllPoints()
        BWdropDownMenu:ClearAllPoints()

        if orientation == "LEFT" then
            B.I.Main.Frame.ButtonLeft("LEFT", "RIGHT")
        else
            B.I.Main.Frame.ButtonRight("RIGHT", "LEFT")
        end
    end

    function B.I.Main.Frame.ButtonLeft(startpoint, direction)

        T.F.mainframe.optionbutton:SetPoint(startpoint, T.F.mainframe, startpoint)
        T.F.mainframe.optionbutton:SetPoint("BOTTOM", T.F.mainframe, "BOTTOM")
        T.F.mainframe.bankbutton:SetPoint(startpoint, T.F.mainframe.optionbutton, direction)
        T.F.mainframe.reagentbutton:SetPoint(startpoint, T.F.mainframe.bankbutton, direction)
        T.F.mainframe.voidbutton:SetPoint(startpoint, T.F.mainframe.reagentbutton, direction)
        T.F.mainframe.guildbutton:SetPoint(startpoint, T.F.mainframe.voidbutton, direction)
        T.F.mainframe.searchbar:SetPoint(startpoint, T.F.mainframe.guildbutton, direction, 10, 0)
        T.F.mainframe.countbutton:SetPoint(startpoint, T.F.mainframe.searchbar, direction)
        T.F.mainframe.search:SetPoint("TOP"..startpoint, T.F.mainframe.searchbar, "BOTTOM"..startpoint, -10, 0)
        T.F.mainframe.search:SetPoint("TOP"..direction, T.F.mainframe.countbutton, "BOTTOM"..direction)

        BWdropDownMenu:SetPoint("TOPLEFT", T.F.mainframe.optionbutton, "BOTTOMLEFT", -15, 24)
    end

    function B.I.Main.Frame.ButtonRight(startpoint, direction)

        T.F.mainframe.optionbutton:SetPoint(startpoint, T.F.mainframe, startpoint)
        T.F.mainframe.optionbutton:SetPoint("BOTTOM", T.F.mainframe, "BOTTOM")
        T.F.mainframe.guildbutton:SetPoint(startpoint, T.F.mainframe.optionbutton, direction)
        T.F.mainframe.voidbutton:SetPoint(startpoint, T.F.mainframe.guildbutton, direction)
        T.F.mainframe.reagentbutton:SetPoint(startpoint, T.F.mainframe.voidbutton, direction)
        T.F.mainframe.bankbutton:SetPoint(startpoint, T.F.mainframe.reagentbutton, direction)
        T.F.mainframe.searchbar:SetPoint(startpoint, T.F.mainframe.bankbutton, direction, -10, 0)
        T.F.mainframe.countbutton:SetPoint(startpoint, T.F.mainframe.searchbar, direction)
        T.F.mainframe.search:SetPoint("TOP"..startpoint, T.F.mainframe.searchbar, "BOTTOM"..startpoint, 10, 0)
        T.F.mainframe.search:SetPoint("TOP"..direction, T.F.mainframe.countbutton, "BOTTOM"..direction)

        BWdropDownMenu:SetPoint("TOPRIGHT", T.F.mainframe.optionbutton, "BOTTOMRIGHT", 0, 24)
    end

    function B.I.Main.Frame:MouseUp(...)
        self:StopMovingOrSizing()
    end

    function B.I.Main.Frame:MouseDown(...)
        self:StartMoving()
    end

    function B.I.Main.Frame.Options()

        T.F.mainframe.optionbutton = B.I.Main.Frame.Button("|cff7E7E7Eo|r", nil, "BWoptionbutton")
        T.F.mainframe.optionbutton:SetHighlightTexture("")
        T.F.mainframe.optionbutton:SetPoint("TOPLEFT", T.F.mainframe, "TOPLEFT")
        T.F.mainframe.optionbutton:SetPoint("BOTTOM", T.F.mainframe, "BOTTOM")
        T.F.mainframe.optionbutton:SetBackdropColor(0.05, 0.05, 0.05, 1)
        T.F.mainframe.optionbutton:RegisterForClicks("LeftButtonUp", "RightButtonUp")
        T.F.mainframe.optionbutton:RegisterForDrag("LeftButton")
        T.F.mainframe.optionbutton:SetScript("OnShow", function(self) B.SetFrameLocation() end)
        T.F.mainframe.optionbutton:SetScript("OnEnter", function(self) self:SetText("|cffffffffo|r")
            if (BWsavedtable and BWsavedtable.helptooltip) or not BWsavedtable then
                GameTooltip:ClearLines()
                GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
                GameTooltip:SetText(T.ACTIVE_TABLE.player.." ("..string.sub(T.ACTIVE_TABLE.faction, 1, 1)..")")
                GameTooltip:AddLine(" ")
                GameTooltip:AddLine(L[10005], 1, 1, 1) -- drag
                GameTooltip:AddLine(L[10006], 1, 1, 1) -- hide
                GameTooltip:AddLine(L[10007], 1, 1, 1) -- menu
                GameTooltip:Show()
            end
        end)
        T.F.mainframe.optionbutton:SetScript("OnLeave", function(self) self:SetText("|cff7E7E7Eo|r"); GameTooltip:Hide() end)
        T.F.mainframe.optionbutton:SetScript("OnDragStart", function(self) self:SetText("|cff0FBCE2o|r"); T.F.mainframe:StartMoving() end)
        T.F.mainframe.optionbutton:SetScript("OnDragStop", function(self) self:SetText("|cff7E7E7Eo|r"); T.F.mainframe:StopMovingOrSizing(); B.SetFrameLocation(); end)
        T.F.mainframe.optionbutton:SetScript("OnClick", function(self, button)
            if button == "RightButton" then
                T.F.mainframe.togglebutton:Show()
                HideDropDownMenu(1)
            else
                ToggleDropDownMenu(1, nil, BWdropDownMenu, "BWoptionbutton", 0, 0)
            end 
        end)

        B.I.OptionsDropDown.Create()
        B.I.Main.Frame.Toggle()
    end

    function B.I.Main.Frame.Bank()

        T.F.mainframe.bankbutton = B.I.Main.Frame.Button("|cffffffffB|r")
        T.F.mainframe.bankbutton:SetPoint("LEFT", T.F.mainframe.optionbutton, "RIGHT")
        T.F.mainframe.bankbutton:SetBackdropColor(0.32, 0.15, 0.64, 1)
        T.F.mainframe.bankbutton:SetScript("OnClick", function(self)
            T.F.mainframe.lastbank = nil
            T.F.mainframe.reagent:Hide()
            T.F.mainframe.void:Hide()
            T.F.mainframe.guild:Hide()
            T.F.mainframe.search:Hide()

            if T.F.mainframe.bank:IsVisible() then
                T.F.mainframe.bank:Hide()
            else
                T.F.mainframe.lastbank = "bank"
                T.F.mainframe.bank:Show()
            end
        end)
    end

    function B.I.Main.Frame.Reagent()

        T.F.mainframe.reagentbutton = B.I.Main.Frame.Button("|cffffffffR|r")
        T.F.mainframe.reagentbutton:SetPoint("LEFT", T.F.mainframe.bankbutton, "RIGHT")
        T.F.mainframe.reagentbutton:SetBackdropColor(0.34, 0.24, 0.49, 1)
        T.F.mainframe.reagentbutton:SetScript("OnClick", function(self)
            T.F.mainframe.lastbank = nil
            T.F.mainframe.bank:Hide()
            T.F.mainframe.void:Hide()
            T.F.mainframe.guild:Hide()
            T.F.mainframe.search:Hide()

            if T.F.mainframe.reagent:IsVisible() then
                T.F.mainframe.reagent:Hide()
            else
                T.F.mainframe.lastbank = "reagent"
                T.F.mainframe.reagent:Show()
            end
        end)
    end

    function B.I.Main.Frame.Void()

        T.F.mainframe.voidbutton = B.I.Main.Frame.Button("|cffffffffV|r")
        T.F.mainframe.voidbutton:SetPoint("LEFT", T.F.mainframe.reagentbutton, "RIGHT")
        T.F.mainframe.voidbutton:SetBackdropColor(0.24, 0.24, 0.49, 1)
        T.F.mainframe.voidbutton:SetScript("OnClick", function(self)
            T.F.mainframe.lastbank = nil
            T.F.mainframe.bank:Hide()
            T.F.mainframe.reagent:Hide()
            T.F.mainframe.guild:Hide()
            T.F.mainframe.search:Hide()

            if T.F.mainframe.void:IsVisible() then
                T.F.mainframe.void:Hide()
            else
                T.F.mainframe.lastbank = "void"
                T.F.mainframe.void:Show()
            end
        end)
    end

    function B.I.Main.Frame.Guild()

        T.F.mainframe.guildbutton = B.I.Main.Frame.Button("|cffffffffG|r")
        T.F.mainframe.guildbutton:SetPoint("LEFT", T.F.mainframe.voidbutton, "RIGHT")
        T.F.mainframe.guildbutton:SetBackdropColor(0.20, 0.17, 0.56, 1)
        T.F.mainframe.guildbutton:SetScript("OnClick", function(self)
            T.F.mainframe.lastbank = nil
            T.F.mainframe.bank:Hide()
            T.F.mainframe.reagent:Hide()
            T.F.mainframe.void:Hide()
            T.F.mainframe.search:Hide()

            if T.F.mainframe.guild:IsVisible() then
                T.F.mainframe.guild:Hide()
            else
                T.F.mainframe.lastbank = "guild"
                T.F.mainframe.guild:Show()
            end
        end)
    end

    function B.I.Main.Frame.Toggle()

        T.F.mainframe.togglebutton = B.I.Main.Frame.Button("|cffE62C2Co|r", T.F.toggleframe)
        T.F.mainframe.togglebutton:SetHighlightTexture("")
        T.F.mainframe.togglebutton:SetAllPoints(T.F.mainframe.optionbutton)
        T.F.mainframe.togglebutton:SetClampedToScreen(true)
        T.F.mainframe.togglebutton:Hide()
        T.F.mainframe.togglebutton:SetBackdropColor(0.07, 0.07, 0.07, 0.8)
        T.F.mainframe.togglebutton:SetBackdropBorderColor(0.15, 0.15, 0.15, 1)
        T.F.mainframe.togglebutton:SetScript("OnEnter", function(self) self:SetText("|cff03C926o|r"); self:SetBackdropBorderColor(0.01, 0.79, 0.15, 1) end)
        T.F.mainframe.togglebutton:SetScript("OnLeave", function(self) self:SetText("|cffE62C2Co|r"); self:SetBackdropBorderColor(0.15, 0.15, 0.15, 1) end)
        T.F.mainframe.togglebutton:SetScript("OnShow", function(self) T.F.mainframe:Hide() end)
        T.F.mainframe.togglebutton:SetScript("OnHide", function(self) T.F.mainframe:Show() end)
        T.F.mainframe.togglebutton:SetScript("OnClick", function(self) self:Hide() end)
    end

    B.I.OptionsDropDown = {}
    function B.I.OptionsDropDown.Create()

        BWdropDownMenu = CreateFrame("Frame", "BWdropDownMenu", T.F.mainframe, "UIDropDownMenuTemplate")
        BWdropDownMenu:Hide()
        UIDropDownMenu_SetWidth(BWdropDownMenu, 130, 4)
        BWdropDownMenu:SetPoint("TOPLEFT", T.F.mainframe.optionbutton, "BOTTOMLEFT", -15, 24)
        UIDropDownMenu_Initialize(BWdropDownMenu, B.I.OptionsDropDown.Initialize) --The virtual
    end

    function B.I.OptionsDropDown:Initialize(level)
        --drop down menues can have sub menues. The value of level determines the drop down sub menu tier
        local level = level or 1
        local x, k, v

        if level == 1 then
            local info = UIDropDownMenu_CreateInfo()
            if BWsavedtable then
                for k, v in pairs(BWsavedtable) do
                    if not T.OPTIONS[k] then B.I.OptionsDropDown.Menu(info, k, k, level) end
                end
            end
        elseif level == 2 then
            local info = UIDropDownMenu_CreateInfo()
            if BWsavedtable then
                for k, v in pairs(BWsavedtable[UIDROPDOWNMENU_MENU_VALUE]) do
                    checked = tostring(T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)) == k and true or false
                    player, faction = k:match"([^.]*).(.*)"
                    B.I.OptionsDropDown.Button(self, info, player.." ("..string.sub(faction, 1, 1)..")", UIDROPDOWNMENU_MENU_VALUE.."."..k, checked, level)
                end
            end
        end
    end

    function B.I.OptionsDropDown.Menu(info, label, value, level)

        info.text = label
        info.value = value
        info.hasArrow = true
        info.checked = false

        UIDropDownMenu_AddButton(info, level)
    end

    function B.I.OptionsDropDown.Button(self, info, label, value, checked, level)

        info.text = label
        info.hasArrow = false
        info.value = value
        info.func = B.dropDownMenuItem_OnClick
        info.owner = self:GetParent()
        info.checked = checked

        UIDropDownMenu_AddButton(info, level)
    end

    function B.I.Main.Frame.Search()

        T.F.mainframe.searchbar = CreateFrame("EditBox", "BW_Searchbar", T.F.mainframe, "BagSearchBoxTemplate")
        T.F.mainframe.searchbar:SetHeight(25)
        T.F.mainframe.searchbar:SetWidth(192)
        T.F.mainframe.searchbar:SetAutoFocus(false)
        T.F.mainframe.searchbar:SetPoint("LEFT", T.F.mainframe.guildbutton, "RIGHT", 10, 0)
        T.F.mainframe.searchbar:SetHyperlinksEnabled(true) 
        T.F.mainframe.searchbar:HookScript("OnEscapePressed", function(self) self:ClearFocus() end)
        T.F.mainframe.searchbar:HookScript("OnTextChanged", function(self)
            text = self:GetText()
            if text == "" then text = false end

            if text ~= SEARCH then

                T.F.mainframe:SetBackdropColor(0.05, 0.05, 0.05, 1)
                T.F.mainframe.countbutton.count = 0
                T.F.mainframe.search.scroll.ScrollBar:SetValue(0)

                B.Search(text)

                if T.F.mainframe.countbutton.count > 0 then
                    T.F.mainframe.countbutton.labelcount:SetText("|cffffffff"..T.F.mainframe.countbutton.count.."|r")
                    T.F.mainframe.countbutton:Show()
                else
                    T.F.mainframe.countbutton:Hide()
                end

                if not text then
                    T.F.mainframe:SetBackdropColor(0, 0, 0, 0.5)
                end
            else
                T.F.mainframe:SetBackdropColor(0, 0, 0, 0.5)
                B.Search(false)
            end
        end)
        B.I.RemoveEdge(T.F.mainframe.searchbar)
        tinsert(ITEM_SEARCHBAR_LIST, "BW_Searchbar")
    end

    function B.I.Main.Frame.Count()

        T.F.mainframe.countbutton = B.I.Main.Frame.Button("|cffffffff|r")
        T.F.mainframe.countbutton:SetWidth(64)
        T.F.mainframe.countbutton:SetPoint("LEFT", T.F.mainframe.searchbar, "RIGHT")
        T.F.mainframe.countbutton:SetBackdropColor(0.05, 0.05, 0.05, 1)
        T.F.mainframe.countbutton:Hide()

        T.F.mainframe.countbutton:SetScript("OnHide", function(self) T.F.mainframe.search:Hide() end)
        T.F.mainframe.countbutton:SetScript("OnClick", function(self)
            T.F.mainframe.bank:Hide()
            T.F.mainframe.reagent:Hide()
            T.F.mainframe.void:Hide()
            T.F.mainframe.guild:Hide()

            if T.F.mainframe.search:IsVisible() then
                T.F.mainframe.search:Hide()
                if T.F.mainframe.lastbank then T.F.mainframe[T.F.mainframe.lastbank]:Show() end
            else
                T.F.mainframe.search:Show()
            end
        end)

        T.F.mainframe.countbutton.labelcount = T.F.mainframe.countbutton:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        T.F.mainframe.countbutton.labelcount:SetFont(T.F.mainframe.countbutton.labelcount:GetFont(), 14) -- Resize string
        T.F.mainframe.countbutton.labelcount:SetJustifyH("CENTER")
        T.F.mainframe.countbutton.labelcount:SetPoint("CENTER")
    end

    B.I.Main.Search = {}
    function B.I.Main.Search.Create()

        T.F.mainframe.search = CreateFrame("Frame", "BWsearch", T.F.mainframe)
        T.F.mainframe.search:SetHeight(200)
        T.F.mainframe.search:SetPoint("TOPLEFT", T.F.mainframe.searchbar, "BOTTOMLEFT", -10, 0)
        T.F.mainframe.search:SetPoint("TOPRIGHT", T.F.mainframe.countbutton, "BOTTOMRIGHT")

        T.F.mainframe.search:Hide()
        T.F.mainframe.search:EnableMouse(true)

        T.F.mainframe.search.enddot = T.F.mainframe.search:CreateTexture(nil, "OVERLAY")
        T.F.mainframe.search.enddot:SetMask("Textures\\MinimapMask")
        T.F.mainframe.search.enddot:SetHeight(5)
        T.F.mainframe.search.enddot:SetWidth(5)
        T.F.mainframe.search.enddot:SetTexture("Interface\\Buttons\\WHITE8x8")
        T.F.mainframe.search.enddot:SetVertexColor(0.7, 0.7, 0.7, 1)
        T.F.mainframe.search.enddot:SetPoint("TOP", T.F.mainframe.search, "BOTTOM", 0, -5)

        B.I.Main.Search.Scroll()
    end

    function B.I.Main.Search.Scroll()

        T.F.mainframe.search.list = CreateFrame("Frame", "FauxScrollFrameBW", T.F.mainframe.search)
        T.F.mainframe.search.list:SetPoint("TOPLEFT", T.F.mainframe.search, "TOPLEFT", 0, -T.FRAMEWHITESPACE)
        T.F.mainframe.search.list:SetPoint("BOTTOMRIGHT", T.F.mainframe.search, "BOTTOMRIGHT")
        T.F.mainframe.search.list:SetWidth(T.F.mainframe.search:GetWidth())

        T.F.mainframe.search.scroll = B.I.ScrollFrame(T.F.mainframe.search.list, "BWScrollFrame",
                                                      B.ScrollbarUpdate, B.I.Main.Search.Button)
    end

    function B.I.Main.Search.Button(index)

        local button = B.I.List.Button(index, T.F.mainframe.search.list)

        button:SetScript("OnEnter", function(self)
            local r, g, b = GetItemQualityColor(self.quality)
            self:SetBackdropBorderColor(r, g, b, 0.5)
            B.I.Button.Enter(self)
        end)
        button:SetScript("OnLeave", function(self) self:SetBackdropBorderColor(0, 0, 0, 0.9); GameTooltip:Hide() end)
        button.leftstring = button:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        button.leftstring:SetJustifyH("LEFT")
        button.leftstring:SetJustifyV("CENTER")
        button.leftstring:SetWordWrap(false)
        button.leftstring:SetPoint("LEFT", button.icon, "RIGHT", 10, 0)
        button.leftstring:SetPoint("RIGHT", button, "RIGHT", -T.F.mainframe.countbutton:GetWidth(), 0)

        button:SetScript("OnClick", function(self) 
            local location = self.location.loc
            local page = tonumber(self.location.num)

            if location == "bank" then
                T.F.mainframe.bankbutton:Click()
            elseif location == "reagent" then
                T.F.mainframe.reagentbutton:Click()
            elseif location == "void" then
                T.F.mainframe.voidbutton:Click()
                T.F.mainframe.void[page].tab:Click()
            elseif location == "guild" then
                T.F.mainframe.guildbutton:Click()
                T.F.mainframe.guild[page].tab:Click()
            end
        end)

       return button
    end

    B.I.Main.Bank = {}
    function B.I.Main.Bank.Create()

        T.F.mainframe.bank = CreateFrame("Frame", "BWbank", T.F.mainframe)
        T.F.mainframe.bank:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        T.F.mainframe.bank:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
        T.F.mainframe.bank:SetFrameStrata("HIGH")
        T.F.mainframe.bank:Hide()

        T.F.mainframe.bank:SetScript("OnShow", function(self) T.F.mainframe.bankbutton:LockHighlight() end)
        T.F.mainframe.bank:SetScript("OnHide", function(self) T.F.mainframe.bankbutton:UnlockHighlight() end)

        B.I.Backdrop(T.F.mainframe.bank)
    end

    function B.I.Main.Bank.Buttons()

        T.F.mainframe.bank.frame = CreateFrame("Frame", nil, T.F.mainframe.bank)
        T.F.mainframe.bank.frame:SetPoint("BOTTOM")
        T.F.mainframe.bank.frame:SetPoint("TOPLEFT", 10, -20)
        T.F.mainframe.bank.frame:SetFrameStrata("HIGH")

        T.F.mainframe.bank.buttons = {}
        local i, button, bankrow, lastbutton

        for i = 1, NUM_BANKGENERIC_SLOTS do

            button = B.I.Button.Create(i, T.F.mainframe.bank.frame)
            button:SetScript("OnEnter", B.I.Button.Enter)
            button:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
            
            if i == 1 then -- First slot
                button:SetPoint("TOPLEFT")
                B.I.bank.num = 1
                bankrow = button

            elseif B.I.bank.num == T.ROW then -- Last column of the row
                button:SetPoint("TOP", bankrow, "BOTTOM", 0, -T.FRAMEWHITESPACE)
                bankrow = button
                B.I.bank.rows = B.I.bank.rows + 1
                B.I.bank.num = 1
            else
                button:SetPoint("TOPLEFT", lastbutton, "TOPRIGHT", T.FRAMEWHITESPACE, 0)
                B.I.bank.num = B.I.bank.num + 1
            end
            T.F.mainframe.bank.buttons[i] = button
            lastbutton = button
        end
        T.F.mainframe.bank:SetHeight(10 + (T.BUTTON_HEIGHT * (B.I.bank.rows + 2)) + (T.FRAMEWHITESPACE * B.I.bank.rows))
        T.F.mainframe.bank:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
    end

    B.I.Main.Reagent = {}
    function B.I.Main.Reagent.Create()

        T.F.mainframe.reagent = CreateFrame("Frame", "BWreagent", T.F.mainframe)
        T.F.mainframe.reagent:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        T.F.mainframe.reagent:SetFrameStrata("HIGH")
        T.F.mainframe.reagent:Hide()

        T.F.mainframe.reagent:SetScript("OnShow", function(self) T.F.mainframe.reagentbutton:LockHighlight() end)
        T.F.mainframe.reagent:SetScript("OnHide", function(self) T.F.mainframe.reagentbutton:UnlockHighlight() end)

        B.I.Backdrop(T.F.mainframe.reagent)
    end

    function B.I.Main.Reagent.Buttons()

        T.F.mainframe.reagent.buttons = {}
        local i, x, rows, num, button, bankrow, lastbutton

        for x = 1, T.REAGENT_PAGES do
            T.F.mainframe.reagent[x] = {}
            T.F.mainframe.reagent[x].buttons = {}
            T.F.mainframe.reagent[x].frame = CreateFrame("Frame", nil, T.F.mainframe.reagent)
            T.F.mainframe.reagent[x].frame:SetPoint("BOTTOM")
            T.F.mainframe.reagent[x].frame:SetPoint("TOPLEFT", 10, -20)
            T.F.mainframe.reagent[x].frame:SetFrameStrata("HIGH")
            if x ~= 1 then T.F.mainframe.reagent[x].frame:Hide() end

            rows = 0
            num = 0

            for i = 1, T.NUM_REAGENT_SLOTS do

                button = B.I.Button.Create(i, T.F.mainframe.reagent[x].frame)
                button:SetScript("OnEnter", B.I.Button.Enter)
                button:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
                
                if i == 1 then -- First slot
                    button:SetPoint("TOPLEFT")
                    rows = rows + 1
                    num = 1
                    bankrow = button

                elseif num == T.ROW then -- Last column of the row
                    button:SetPoint("TOP", bankrow, "BOTTOM", 0, -T.FRAMEWHITESPACE)
                    rows = rows + 1
                    num = 1
                    bankrow = button
                else
                    button:SetPoint("TOPLEFT", lastbutton, "TOPRIGHT", T.FRAMEWHITESPACE, 0)
                    num = num + 1
                end
                T.F.mainframe.reagent[x].buttons[i] = button
                lastbutton = button
            end
        end
        T.F.mainframe.reagent:SetHeight(10 + (T.BUTTON_HEIGHT * (rows + 1)) + (T.FRAMEWHITESPACE * rows))
        T.F.mainframe.reagent:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
    end

    B.I.Main.Void = {}
    function B.I.Main.Void.Create()

        T.F.mainframe.void = CreateFrame("Frame", "BWvoid", T.F.mainframe)
        T.F.mainframe.void:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        T.F.mainframe.void:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
        T.F.mainframe.void:SetFrameStrata("HIGH")
        T.F.mainframe.void.lasttab = 1
        T.F.mainframe.void:Hide()

        T.F.mainframe.void:SetScript("OnShow", function(self) T.F.mainframe.voidbutton:LockHighlight() end)
        T.F.mainframe.void:SetScript("OnHide", function(self) T.F.mainframe.voidbutton:UnlockHighlight() end)

        B.I.Backdrop(T.F.mainframe.void)
        local x, lasttab

        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x] = {}
            T.F.mainframe.void[x].tab = B.I.Tab.Create(T.F.mainframe.void)
            T.F.mainframe.void[x].tab.index = x
            T.F.mainframe.void[x].tab:SetScript("OnClick", function(self)
                T.F.mainframe.void[T.F.mainframe.void.lasttab].frame:Hide()
                T.F.mainframe.void[T.F.mainframe.void.lasttab].tab:SetBackdropBorderColor(0, 0, 0, 1)
                self:SetBackdropBorderColor(1, 1, 1, 1)
                T.F.mainframe.void[self.index].frame:Show()
                T.F.mainframe.void.lasttab = self.index
            end)

            if x == 1 then
                T.F.mainframe.void[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
                T.F.mainframe.void[x].tab:SetNormalTexture("Interface\\Icons\\INV_Enchant_EssenceCosmicGreater")
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", T.F.mainframe.void, "TOPRIGHT", 0, -1)
            else
                T.F.mainframe.void[x].tab:SetNormalTexture("Interface\\Icons\\INV_Enchant_EssenceArcaneLarge")
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.void[x].tab
        end
    end

    function B.I.Main.Void.Buttons()

        local i, x, button, lastrow, rows, num

        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x].buttons = {}
            T.F.mainframe.void[x].frame = CreateFrame("Frame", nil, T.F.mainframe.void)
            T.F.mainframe.void[x].frame:SetPoint("BOTTOM")
            T.F.mainframe.void[x].frame:SetPoint("TOPLEFT", 10, -20)
            T.F.mainframe.void[x].frame:SetFrameStrata("HIGH")
            if x ~= 1 then T.F.mainframe.void[x].frame:Hide() end

            rows = 0
            num = 0

            for i = 1, T.NUM_VOID_SLOTS do

                button = B.I.Button.Create(i, T.F.mainframe.void[x].frame)
                button:SetScript("OnEnter", B.I.Button.Enter)
                button:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
                
                if i == 1 then -- First slot
                    button:SetPoint("TOPLEFT")
                    rows = rows + 1
                    num = 1
                    lastrow = button

                elseif num == T.ROW then -- Last column of the row
                    button:SetPoint("TOP", lastrow, "BOTTOM", 0, -T.FRAMEWHITESPACE)
                    rows = rows + 1
                    num = 1
                    lastrow = button
                else
                    button:SetPoint("TOPLEFT", lastbutton, "TOPRIGHT", T.FRAMEWHITESPACE, 0)
                    num = num + 1
                end
                T.F.mainframe.void[x].buttons[i] = button
                lastbutton = button
            end
        end
        T.F.mainframe.void:SetHeight(10 + (T.BUTTON_HEIGHT * (rows + 1)) + (T.FRAMEWHITESPACE * rows))
        T.F.mainframe.void:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
    end

    B.I.Main.Guild = {}
    function B.I.Main.Guild.Create()

        T.F.mainframe.guild = CreateFrame("Frame", "BWguild", T.F.mainframe)
        T.F.mainframe.guild:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        T.F.mainframe.guild:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
        T.F.mainframe.guild:SetFrameStrata("HIGH")
        T.F.mainframe.guild.lasttab = 1
        T.F.mainframe.guild:Hide()

        T.F.mainframe.guild:SetScript("OnShow", function(self) T.F.mainframe.guildbutton:LockHighlight() end)
        T.F.mainframe.guild:SetScript("OnHide", function(self) T.F.mainframe.guildbutton:UnlockHighlight() end)

        B.I.Backdrop(T.F.mainframe.guild)
        local x, lasttab

        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x] = {}
            T.F.mainframe.guild[x].tab = B.I.Tab.Create(T.F.mainframe.guild)
            T.F.mainframe.guild[x].tab.index = x
            T.F.mainframe.guild[x].tab:SetScript("OnClick", function(self)
                T.F.mainframe.guild[T.F.mainframe.guild.lasttab].frame:Hide()
                T.F.mainframe.guild[T.F.mainframe.guild.lasttab].tab:SetBackdropBorderColor(0, 0, 0, 1)
                self:SetBackdropBorderColor(1, 1, 1, 1)
                T.F.mainframe.guild[self.index].frame:Show()
                T.F.mainframe.guild.lasttab = self.index
            end)
            T.F.mainframe.guild[x].tab:SetScript("OnEnter", function(self) 
                GameTooltip:SetOwner(self, "ANCHOR_RIGHT", -10, -10)
                if self.name then
                    GameTooltip:SetText(self.name)
                end
            end)
            T.F.mainframe.guild[x].tab:SetScript("OnLeave", function(self) GameTooltip:Hide() end)

            if x == 1 then
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", T.F.mainframe.guild, "TOPRIGHT", 0, -1)
                T.F.mainframe.guild[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
            else
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.guild[x].tab
        end
    end

    function B.I.Main.Guild.Buttons()

        local i, x, button, lastrow, rows, num

        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x].buttons = {}
            T.F.mainframe.guild[x].frame = CreateFrame("Frame", nil, T.F.mainframe.guild)
            T.F.mainframe.guild[x].frame:SetPoint("BOTTOM")
            T.F.mainframe.guild[x].frame:SetPoint("TOPLEFT", 10, -20)
            T.F.mainframe.guild[x].frame:SetFrameStrata("HIGH")
            if x ~= 1 then
                T.F.mainframe.guild[x].frame:Hide()
            end

            rows = 0
            num = 0

            for i = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                button = B.I.Button.Create(i, T.F.mainframe.guild[x].frame)
                button:SetScript("OnEnter", B.I.Button.Enter)
                button:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
                
                if i == 1 then -- First slot
                    button:SetPoint("TOPLEFT")
                    rows = rows + 1
                    num = 1
                    lastrow = button

                elseif num == T.ROW then -- Last column of the row
                    button:SetPoint("TOP", lastrow, "BOTTOM", 0, -T.FRAMEWHITESPACE)
                    rows = rows + 1
                    num = 1
                    lastrow = button
                else
                    button:SetPoint("TOPLEFT", lastbutton, "TOPRIGHT", T.FRAMEWHITESPACE, 0)
                    num = num + 1
                end
                T.F.mainframe.guild[x].buttons[i] = button
                lastbutton = button
            end
        end
        T.F.mainframe.guild:SetHeight(10 + (T.BUTTON_HEIGHT * (rows + 1)) + (T.FRAMEWHITESPACE * rows))
        T.F.mainframe.guild:SetWidth(20 + (T.BUTTON_HEIGHT * T.ROW) + (T.FRAMEWHITESPACE * (T.ROW - 1)))
    end

    local button_border = function(button, alpha, r, g, b)

            if button.quest then
                button:SetBackdropBorderColor(1, 1, 0, alpha)
            elseif button.quality and button.quality > 1 then
                r, g, b = GetItemQualityColor(button.quality)
                button:SetBackdropBorderColor(r, g, b, alpha)
            else
                button:SetBackdropBorderColor(r or 0, g or 0, b or 0, alpha)
            end
    end

    B.I.Button = {}
    function B.I.Button.Create(index, parent)

        local button = CreateFrame("Button", nil, parent)

        button.buttonnumber = index
        button:SetHeight(T.BUTTON_HEIGHT)
        button:SetWidth(T.BUTTON_HEIGHT)
        button:SetFrameStrata("HIGH")

        button:SetBackdrop({bgFile = "Interface\\Buttons\\WHITE8x8", edgeFile = "Interface\\Buttons\\WHITE8x8", edgeSize = 1})
        button:SetBackdropColor(0, 0, 0, 0.2)
        button:SetBackdropBorderColor(0, 0, 0, 1)

        local icon = button:CreateTexture(nil, "ARTWORK", -1)

        icon:SetPoint("CENTER")
        icon:SetHeight(T.BUTTON_HEIGHT - 2)
        icon:SetWidth(T.BUTTON_HEIGHT - 2)
        icon:SetTexture("Interface\\AddOns\\BanksWatch\\media\\gloss")
        icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        button.icon = icon

        local search = CreateFrame("Frame", nil, button)
        search:SetBackdrop({bgFile = "Interface\\Buttons\\WHITE8x8"})
        search:SetBackdropColor(0, 0, 0, 0.8)
        search:SetAllPoints(icon)
        search:SetFrameStrata("HIGH")
        search:SetFrameLevel(6)
        search:Hide()
        search:HookScript("OnShow", function(self) button_border(self:GetParent(), 0.4) end)-- search activated, but not a match
        search:HookScript("OnHide", function(self) button_border(self:GetParent(), 1) end) -- reset button to normal
        button.search = search

        local highlight = button:CreateTexture(nil, "HIGHLIGHT")

        highlight:SetAllPoints()
        highlight:SetTexture("Interface\\Buttons\\ButtonHilight-Square")
        highlight:SetVertexColor(0.949, 0.847, 0.745, 1)
        button:SetHighlightTexture(highlight)

        button.text = button:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        button.text:SetTextColor(1, 1, 1)
        button.text:SetFont(button.text:GetFont(), 14, "OUTLINE")
        button.text:SetJustifyH("LEFT")
        button.text:SetJustifyV("CENTER")
        button.text:SetWordWrap(false)
        button.text:SetPoint("RIGHT")
        button.text:SetPoint("BOTTOM", 0, 3)

        button.ilvl = button:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        button.ilvl:SetFont(button.ilvl:GetFont(), 13, "OUTLINE")
        button.ilvl:SetJustifyH("LEFT")
        button.ilvl:SetJustifyV("CENTER")
        button.ilvl:SetWordWrap(false)
        button.ilvl:SetPoint("LEFT", 1)
        button.ilvl:SetPoint("BOTTOM", 0, 2)

        button:HookScript("OnClick", function(self)
            if IsShiftKeyDown() then
                if T.F.mainframe.searchbar:HasFocus() then T.F.mainframe.searchbar:SetText(self.name)
                else SetItemRef(self.link, self.link, "LeftButton")
                end
            end
        end)

        return button
    end

    B.I.List = {}
    function B.I.List.Button(index, parent)

        local button = CreateFrame("Button", nil, parent)

        button.buttonnumber = index
        button:SetHeight(32)
        button:SetWidth(parent:GetWidth())
        button:SetPoint("TOP")
        button:SetPoint("LEFT")
        button:SetPoint("RIGHT")
        button:SetBackdrop({bgFile = "Interface\\Buttons\\WHITE8x8", edgeFile = "Interface\\Buttons\\WHITE8x8", edgeSize = 2})
        button:SetBackdropColor(0, 0, 0, 0.7)
        button:SetBackdropBorderColor(0, 0, 0, 0.9)

        local highlight = button:CreateTexture(nil, "HIGHLIGHT")

        highlight:SetHeight(button:GetHeight() - 2)
        highlight:SetWidth(button:GetWidth() - 2)
        highlight:SetPoint("CENTER")
        highlight:SetTexture("Interface\\Buttons\\WHITE8x8")
        highlight:SetVertexColor(0.7, 0.7, 0.7, 0.05)
        button:SetHighlightTexture(highlight)

        local icon = button:CreateTexture(nil, "ARTWORK", -1)

        icon:SetHeight(28)
        icon:SetWidth(28)
        icon:SetTexture("Interface\\AddOns\\BanksWatch\\media\\gloss")
        icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        icon:SetPoint("TOPLEFT", 2, -2)
        button.icon = icon

        button.text = button:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        button.text:SetTextColor(1, 1, 1)
        button.text:SetFont(button.text:GetFont(), 12, "OUTLINE")
        button.text:SetJustifyH("LEFT")
        button.text:SetJustifyV("CENTER")
        button.text:SetWordWrap(false)
        button.text:SetPoint("RIGHT", icon)
        button.text:SetPoint("BOTTOM", icon, 0, 3)

        button.ilvl = button:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        button.ilvl:SetFont(button.ilvl:GetFont(), 12, "OUTLINE")
        button.ilvl:SetJustifyH("LEFT")
        button.ilvl:SetJustifyV("CENTER")
        button.ilvl:SetWordWrap(false)
        button.ilvl:SetPoint("LEFT", icon, 1)
        button.ilvl:SetPoint("BOTTOM", icon, 0, 2)

        button.leftstring = button:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        button.leftstring:SetJustifyH("LEFT")
        button.leftstring:SetJustifyV("CENTER")
        button.leftstring:SetWordWrap(false)

        button.rightstring = button:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        button.rightstring:SetJustifyH("LEFT")
        button.rightstring:SetJustifyV("CENTER")
        button.rightstring:SetPoint("RIGHT", -10, 0)
        button.rightstring:SetWordWrap(false)
        button.rightstring:SetTextColor(1, 1, 1)

        button.rightstring.bg = button:CreateTexture(nil, "ARTWORK")
        button.rightstring.bg:SetPoint("RIGHT", -2, 0)
        button.rightstring.bg:SetHeight(button:GetHeight() - 4)
        button.rightstring.bg:SetWidth(45)
        button.rightstring.bg:SetTexture("Interface\\Buttons\\WHITE8x8")
        button.rightstring.bg:SetVertexColor(0.05, 0.05, 0.05, 1)

        return button
    end

    B.I.Tab = {}
    function B.I.Tab.Create(parent)

        local button = CreateFrame("Button", nil, parent)
        button:SetHeight(32)
        button:SetWidth(33)
        button:SetFrameStrata("HIGH")

        button:SetBackdrop({bgFile = "Interface\\Buttons\\WHITE8x8", edgeFile = "Interface\\Buttons\\WHITE8x8", edgeSize = 1})
        button:SetBackdropColor(0, 0, 0, 0.2)
        button:SetBackdropBorderColor(0, 0, 0, 1)

        local normal = button:CreateTexture(nil, "BACKGROUND")

        normal:SetPoint("CENTER")
        normal:SetHeight(button:GetHeight() - 2)
        normal:SetWidth(button:GetWidth() - 2)
        normal:SetTexture("Interface\\AddOns\\BanksWatch\\media\\gloss")
        normal:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        button:SetNormalTexture(normal)

        local search = button:CreateTexture(nil, "TOOLTIP")

        search:SetAllPoints(normal)
        search:SetTexture("Interface\\Buttons\\WHITE8x8")
        search:SetVertexColor(0, 0, 0, 0.8)
        search:Hide()
        button.search = search

        return button
    end

    function B.I.Button:Enter(...)

        if self.link then
            GameTooltip:SetOwner(self, "ANCHOR_NONE")
            GameTooltip:SetHyperlink(self.link)
            GameTooltip:ClearAllPoints()
            GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10, -20)
            if self.quest then
                GameTooltip:SetBackdropBorderColor(1, 1, 0)
            elseif self.quality then
                local r, g, b = GetItemQualityColor(self.quality)
                GameTooltip:SetBackdropBorderColor(r, g, b)
            end
            GameTooltip:Show()
        end
    end

    function B.I.Backdrop(frame)

        frame:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
                            edgeFile = "Interface\\ChatFrame\\ChatFrameBackground",
                            tile = false, tileSize = 32, edgeSize = 1,
                            insets = { left = 0, right = 0, top = 0, bottom = 0 }
        })
        frame:SetBackdropBorderColor(0, 0, 0, 1)
        frame:SetBackdropColor(0, 0, 0, 0.4)
        frame:EnableMouse(true)
    end

    B.I.GradientButton = {}
    function B.I.GradientButton.Create(button, orientation)

        button:SetNormalTexture("")
        button:SetHighlightTexture("")
        button:SetPushedTexture("")
        button:SetDisabledTexture("")

        B.I.RemoveEdge(button)

        button:SetBackdrop({bgFile = AS_backdrop,
                            edgeFile = AS_backdrop,
                            edgeSize = 1,
                            insets = { left = 0, right = 0, top = 0, bottom = 0 }
        })
        button:SetBackdropColor(0, 0, 0, 0)
        button:SetBackdropBorderColor(0, 0, 0, 1)

        local gradient = button:CreateTexture(nil, "BACKGROUND")

        gradient:SetColorTexture(1, 1, 1)
        gradient:SetAllPoints()
        gradient:SetGradient(orientation, 0, 0, 0, 0.15, 0.15, 0.15)
    end

    function B.I.RemoveEdge(frame)

        local name = frame:GetName()
        local left = frame.Left or name and _G[name.."Left"] or nil
        local middle = frame.Middle or name and (_G[name.."Middle"] or _G[name.."Mid"]) or nil
        local right = frame.Right or name and _G[name.."Right"] or nil

        if left then left:Hide() end
        if middle then middle:Hide() end
        if right then right:Hide() end
    end

    function B.I.AnchorBagsTopLeft()
        local x, lasttab
        -- DEFAULT LAYOUT
        T.F.mainframe.layout = 1
        B.I.Main.Frame.ButtonOrder("LEFT")
        -- bank
        T.F.mainframe.bank:ClearAllPoints()
        T.F.mainframe.bank:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.bankbutton, "BOTTOM")

        -- reagent
        T.F.mainframe.reagent:ClearAllPoints()
        T.F.mainframe.reagent:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.reagentbutton, "BOTTOM")

        -- void
        T.F.mainframe.void:ClearAllPoints()
        T.F.mainframe.void:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.voidbutton, "BOTTOM")
        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", T.F.mainframe.void, "TOPRIGHT", 0, -1)
            else
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.void[x].tab
        end

        -- guild
        T.F.mainframe.guild:ClearAllPoints()
        T.F.mainframe.guild:SetPoint("TOPLEFT", T.F.mainframe, "BOTTOMLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.guildbutton, "BOTTOM")
        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", T.F.mainframe.guild, "TOPRIGHT", 0, -1)
                T.F.mainframe.guild[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
            else
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.guild[x].tab
        end

        -- search
        T.F.mainframe.search:ClearAllPoints()
        T.F.mainframe.search:SetPoint("TOPLEFT", T.F.mainframe.searchbar, "BOTTOMLEFT", -10, 0)
        T.F.mainframe.search:SetPoint("TOPRIGHT", T.F.mainframe.countbutton, "BOTTOMRIGHT")
        T.F.mainframe.search.enddot:SetPoint("TOP", T.F.mainframe.search, "BOTTOM", 0, -5)
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.countbutton, "BOTTOM")

        local parent = T.F.mainframe.search.list
        for i = 1, B.RowsToFit() do
            currentrow = parent.buttons[i]
            currentrow:ClearAllPoints()
            if i == 1 then
                currentrow:SetPoint("TOP")
            else
                currentrow:SetPoint("TOP", previousrow, "BOTTOM", 0, -1)
            end
            currentrow:Show()
            previousrow = currentrow
        end
    end

    function B.I.AnchorBagsBottomLeft()

        T.F.mainframe.layout = 2
        B.I.Main.Frame.ButtonOrder("LEFT")
        -- bank
        T.F.mainframe.bank:ClearAllPoints()
        T.F.mainframe.bank:SetPoint("BOTTOMLEFT", T.F.mainframe, "TOPLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.bankbutton, "TOP")

        -- reagent
        T.F.mainframe.reagent:ClearAllPoints()
        T.F.mainframe.reagent:SetPoint("BOTTOMLEFT", T.F.mainframe, "TOPLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.reagentbutton, "TOP")

        -- void
        T.F.mainframe.void:ClearAllPoints()
        T.F.mainframe.void:SetPoint("BOTTOMLEFT", T.F.mainframe, "TOPLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.voidbutton, "TOP")
        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", T.F.mainframe.void, "TOPRIGHT", 0, -1)
            else
                T.F.mainframe.void[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.void[x].tab
        end

        -- guild
        T.F.mainframe.guild:ClearAllPoints()
        T.F.mainframe.guild:SetPoint("BOTTOMLEFT", T.F.mainframe, "TOPLEFT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.guildbutton, "TOP")
        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", T.F.mainframe.guild, "TOPRIGHT", 0, -1)
                T.F.mainframe.guild[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
            else
                T.F.mainframe.guild[x].tab:SetPoint("TOPLEFT", lasttab, "BOTTOMLEFT", 0, -4)
            end
            lasttab = T.F.mainframe.guild[x].tab
        end

        -- search
        T.F.mainframe.search:ClearAllPoints()
        T.F.mainframe.search:SetPoint("BOTTOMLEFT", T.F.mainframe.searchbar, "TOPLEFT", -10, 0)
        T.F.mainframe.search:SetPoint("BOTTOMRIGHT", T.F.mainframe.countbutton, "TOPRIGHT")
        T.F.mainframe.search.enddot:SetPoint("TOP", T.F.mainframe.search, "TOP", 0, 5)
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.countbutton, "TOP")

        local parent = T.F.mainframe.search.list
        for i = 1, B.RowsToFit() do
            currentrow = parent.buttons[i]
            currentrow:ClearAllPoints()
            if i == 1 then
                currentrow:SetPoint("BOTTOM")
            else
                currentrow:SetPoint("BOTTOM", previousrow, "TOP", 0, 1)
            end
            currentrow:Show()
            previousrow = currentrow
        end
    end

    function B.I.AnchorBagsTopRight()

        T.F.mainframe.layout = 3
        B.I.Main.Frame.ButtonOrder("RIGHT")
        -- bank
        T.F.mainframe.bank:ClearAllPoints()
        T.F.mainframe.bank:SetPoint("TOPRIGHT", T.F.mainframe, "BOTTOMRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.bankbutton, "BOTTOM")

        -- reagent
        T.F.mainframe.reagent:ClearAllPoints()
        T.F.mainframe.reagent:SetPoint("TOPRIGHT", T.F.mainframe, "BOTTOMRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.reagentbutton, "BOTTOM")

        -- void
        T.F.mainframe.void:ClearAllPoints()
        T.F.mainframe.void:SetPoint("TOPRIGHT", T.F.mainframe, "BOTTOMRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.voidbutton, "BOTTOM")
        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.void[x].tab:SetPoint("TOPRIGHT", T.F.mainframe.void, "TOPLEFT", 0, -1)
            else
                T.F.mainframe.void[x].tab:SetPoint("TOPRIGHT", lasttab, "BOTTOMRIGHT", 0, -4)
            end
            lasttab = T.F.mainframe.void[x].tab
        end

        -- guild
        T.F.mainframe.guild:ClearAllPoints()
        T.F.mainframe.guild:SetPoint("TOPRIGHT", T.F.mainframe, "BOTTOMRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.guildbutton, "BOTTOM")
        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.guild[x].tab:SetPoint("TOPRIGHT", T.F.mainframe.guild, "TOPLEFT", 0, -1)
                T.F.mainframe.guild[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
            else
                T.F.mainframe.guild[x].tab:SetPoint("TOPRIGHT", lasttab, "BOTTOMRIGHT", 0, -4)
            end
            lasttab = T.F.mainframe.guild[x].tab
        end

        -- search
        T.F.mainframe.search:ClearAllPoints()
        T.F.mainframe.search:SetPoint("TOPRIGHT", T.F.mainframe.searchbar, "BOTTOMRIGHT", 10, 0)
        T.F.mainframe.search:SetPoint("TOPLEFT", T.F.mainframe.countbutton, "BOTTOMLEFT")
        T.F.mainframe.search.enddot:SetPoint("TOP", T.F.mainframe.search, "BOTTOM", 0, -5)
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.countbutton, "BOTTOM")

        local parent = T.F.mainframe.search.list
        for i = 1, B.RowsToFit() do
            currentrow = parent.buttons[i]
            currentrow:ClearAllPoints()
            if i == 1 then
                currentrow:SetPoint("TOP")
            else
                currentrow:SetPoint("TOP", previousrow, "BOTTOM", 0, -1)
            end
            currentrow:Show()
            previousrow = currentrow
        end
    end

    function B.I.AnchorBagsBottomRight()

        T.F.mainframe.layout = 4
        B.I.Main.Frame.ButtonOrder("RIGHT")
        -- bank
        T.F.mainframe.bank:ClearAllPoints()
        T.F.mainframe.bank:SetPoint("BOTTOMRIGHT", T.F.mainframe, "TOPRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.bankbutton, "TOP")

        -- reagent
        T.F.mainframe.reagent:ClearAllPoints()
        T.F.mainframe.reagent:SetPoint("BOTTOMRIGHT", T.F.mainframe, "TOPRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.reagentbutton, "TOP")

        -- void
        T.F.mainframe.void:ClearAllPoints()
        T.F.mainframe.void:SetPoint("BOTTOMRIGHT", T.F.mainframe, "TOPRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.voidbutton, "TOP")
        for x = 1, T.VOID_PAGES do
            T.F.mainframe.void[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.void[x].tab:SetPoint("TOPRIGHT", T.F.mainframe.void, "TOPLEFT", 0, -1)
            else
                T.F.mainframe.void[x].tab:SetPoint("TOPRIGHT", lasttab, "BOTTOMRIGHT", 0, -4)
            end
            lasttab = T.F.mainframe.void[x].tab
        end

        -- guild
        T.F.mainframe.guild:ClearAllPoints()
        T.F.mainframe.guild:SetPoint("BOTTOMRIGHT", T.F.mainframe, "TOPRIGHT")
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.guildbutton, "TOP")
        for x = 1, MAX_GUILDBANK_TABS do
            T.F.mainframe.guild[x].tab:ClearAllPoints()
            if x == 1 then
                T.F.mainframe.guild[x].tab:SetPoint("TOPRIGHT", T.F.mainframe.guild, "TOPLEFT", 0, -1)
                T.F.mainframe.guild[x].tab:SetBackdropBorderColor(1, 1, 1, 1)
            else
                T.F.mainframe.guild[x].tab:SetPoint("TOPRIGHT", lasttab, "BOTTOMRIGHT", 0, -4)
            end
            lasttab = T.F.mainframe.guild[x].tab
        end

        -- search
        T.F.mainframe.search:ClearAllPoints()
        T.F.mainframe.search:SetPoint("BOTTOMRIGHT", T.F.mainframe.searchbar, "TOPRIGHT", 10, 0)
        T.F.mainframe.search:SetPoint("BOTTOMLEFT", T.F.mainframe.countbutton, "TOPLEFT")
        T.F.mainframe.search.enddot:SetPoint("TOP", T.F.mainframe.search, "TOP", 0, 5)
        B.I.Main.Frame.ButtonHighlight(T.F.mainframe.countbutton, "TOP")

        local parent = T.F.mainframe.search.list
        for i = 1, B.RowsToFit() do
            currentrow = parent.buttons[i]
            currentrow:ClearAllPoints()
            if i == 1 then
                currentrow:SetPoint("BOTTOM")
            else
                currentrow:SetPoint("BOTTOM", previousrow, "TOP", 0, 1)
            end
            currentrow:Show()
            previousrow = currentrow
        end
    end

    function B.I.ScrollFrame(parent, name, onUpdate, buttons)

        local scrollframe = B.I.Scroll.Create(parent, name, onUpdate)
        local i, currentrow, previousrow

        parent.buttons = {}

        for i = 1, B.RowsToFit() do
            parent.buttons[i] = buttons(i)
            currentrow = parent.buttons[i]
            if i == 1 then
                currentrow:SetPoint("TOP")
            else
                currentrow:SetPoint("TOP", previousrow, "BOTTOM", 0, -1)
            end
            currentrow:Show()
            previousrow = currentrow
        end

        return scrollframe
    end

    B.I.Scroll = {}
    function B.I.Scroll.Create(parent, name, onUpdate)

        --[[note the anchors: the area of the scrollframe is the scrollable area
            that intercepts mousewheel to scroll. it does not include the scrollbar,
            which is anchored off the right ]]

        local scroll = CreateFrame("ScrollFrame", name, parent, "FauxScrollFrameTemplate")
        scroll:SetPoint("TOPLEFT", T.F.mainframe.search, "TOPLEFT")
        scroll:SetPoint("BOTTOMRIGHT", T.F.mainframe.search, "BOTTOMRIGHT")
        scroll.onUpdate = onUpdate

        scroll:SetScript("OnShow", onUpdate)
        scroll:SetScript("OnVerticalScroll", B.I.Scroll.VerticalScroll)
        scroll.ScrollBar:Hide()

        return scroll
    end

    function B.I.Scroll:VerticalScroll(offset)
        FauxScrollFrame_OnVerticalScroll(self, offset, 32, self.onUpdate)
    end
